import grapa.layers.gpre4
import grapa.layers.gcomextractdiag
import grapa.layers.gaddzeros
import grapa.layers.gkeepmatcut
import grapa.layers.gcomfullyconnected
import grapa.layers.gliam
import grapa.layers.gsym
import grapa.layers.glam
import grapa.layers.gbuilder
import grapa.layers.gcomgraphand
import grapa.layers.gcomdensediverge
import grapa.layers.gremoveparam
import grapa.layers.gcomgraphand2
import grapa.layers.glom
import grapa.layers.gcomgraphcutter
import grapa.layers.gltk
import grapa.layers.gfeat
import grapa.layers.gpoolgrowth
import grapa.layers.gltknd
import grapa.layers.ggraphstract
import grapa.layers.gpre2
import grapa.layers.gperm
import grapa.layers.gaddbias
import grapa.layers.gcomdex
import grapa.layers.gtlbuilder
import grapa.layers.gcomgraphcombinations
import grapa.layers.gltrivmlp
import grapa.layers.gcomdiagraph
import grapa.layers.gcomreopool
import grapa.layers.gcomdepoollg
import grapa.layers.gcomgraphfrom2param
import grapa.layers.gpartinorm
import grapa.layers.gpre1
import grapa.layers.glim
import grapa.layers.gl
import grapa.layers.glmlp
import grapa.layers.gchooseparam
import grapa.layers.gfromparam
import grapa.layers.glkeep
import grapa.layers.gcomdensemerge
import grapa.layers.kron
import grapa.layers.gshuffle
import grapa.layers.gcutparam
import grapa.layers.glcreate
import grapa.layers.gcomdepool
import grapa.layers.gcomparastract
import grapa.layers.gcompoolmerge
import grapa.layers.gfeatkeep
import grapa.layers.gcomdepoolplus
import grapa.layers.gkeepcutter
import grapa.layers.glbuilder
import grapa.layers.gortho
import grapa.layers.gvaluation
import grapa.layers.gcomparamcombinations
import grapa.layers.ggoparam
import grapa.layers.gcomgraphfromparam
import grapa.layers.gbrokengrowth
import grapa.layers.glm
import grapa.layers.gcutter
import grapa.layers.glacreate
import grapa.layers.gcomjpool
import grapa.layers.gtopk
import grapa.layers.gssort
import grapa.layers.gcomgraphlevel
import grapa.layers.gadd1
import grapa.layers.gecutter
import grapa.layers.gpre3
import grapa.layers.gcompool
import grapa.layers.ghealparam
import grapa.layers.gpre5
import grapa.layers.gcomgraphrepeat
import grapa.layers.gkeepbuilder
import grapa.layers.gtbuilder
import grapa.layers.gpool
import grapa.layers.gmultiply
import grapa.layers.gcomgpool
import grapa.layers.gmake1graph
import grapa.layers.gcomparamlevel



import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gremoveparam(Layer):
  def __init__(self,gs=50,inn=30,out=20,**kwargs):
    assert inn>=out
    self.gs=gs
    self.inn=inn
    self.out=out
    super(gremoveparam,self).__init__(**kwargs)


  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]
    return x[:,:self.gs,:self.out] 
 

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[-1]==self.inn
    assert input_shape[-2]==self.gs
    return tuple([input_shape[0],self.gs,self.out])    

  def get_config(self):
    mi={"inn":self.inn,"gs":self.gs,"out":self.out}
    th=super(gremoveparam,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gremoveparam(**config)
















from tensorflow.python.ops import array_ops

def kronecker_product(mat1, mat2):
  """Computes the Kronecker product two matrices."""
  m1, n1 = mat1.get_shape().as_list()
  mat1_rsh = array_ops.reshape(mat1, [m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [m1 * m2, n1 * n2])
def kronecker_product_b1(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1)=1."""
  m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1,m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kronecker_product_b2(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat2)=1."""
  m1, n1 = mat1.get_shape().as_list()
  mat1_rsh = array_ops.reshape(mat1, [m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()[1:]
  mat2_rsh = array_ops.reshape(mat2, [-1,1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kronecker_product_bb(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1/2)=1."""
  m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1, m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()[1:]
  mat2_rsh = array_ops.reshape(mat2, [-1, 1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kronecker_product_b1fx1(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1)=1."""
  f1, m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1,f1,m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,f1,m1 * m2, n1 * n2])



#from (or based on)
#https://github.com/tensorflow/tensorflow/blob/r1.8/tensorflow/contrib/kfac/python/ops/utils.py




import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class glacreate(Layer):
  def __init__(self,gs=20,param=40,a=2,**kwargs):
    self.gs=gs
    self.param=param
    self.a=a

    super(glacreate,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"a":self.a}
    th=super(glacreate,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return glacreate(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    #print("!",x[0].shape,x[1].shape)
    mat=x[0]
    val=x[1]

    rmat=K.reshape(mat,(-1,self.gs,self.gs))
    rval=K.reshape(val,(-1,self.gs,self.param))
    
    rwei=K.batch_dot(rmat,rval)
    wei=K.reshape(rwei,(-1,self.a,self.gs,self.param)) 



    ret=K.concatenate((val,wei),axis=-1)

    
    return ret


    
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==4
    assert pshape[-3]==self.a
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs

    shape=list(input_shape[1])
    assert len(shape)==4
    assert shape[-3]==self.a
    assert shape[-1]==self.param
    assert shape[-2]==self.gs

    return tuple([shape[0],self.a,self.gs,self.param*2])
















import numpy as np
import math
import sys

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001

def phieta_init(shape,dtype=None):
  rel=np.zeros(shape)
  if len(rel)>=4:
    rel[4,0]=1.0
  if len(rel)>=5:
    rel[5,0]=1.0
  return K.constant(rel,dtype=dtype)


class gtopk(Layer):
  def __init__(self,gs=30,param=20,k=4,free=0,learnable=True,self_interaction=False,metrik_init=keras.initializers.TruncatedNormal(mean=0.0,stddev=0.05),self_interaction_const=100.0,numericalC=10000,emptyconst=100000000.0,flag=7,**kwargs):
    
    print("initilized a new gtopk with gs=",gs)
    self.gs=gs
    self.k=k
    self.param=param
    self.free=free
    self.learnable=learnable
    self.self_interaction=self_interaction
    self.self_interaction_const=self_interaction_const
    self.numericalC=numericalC
    self.metrik_init=metrik_init
    self.flag=flag
    self.emptyconst=emptyconst


    #print(kwargs)
    #print((gs,param))
    #exit()

    super(gtopk,self).__init__(input_shape=(gs,param))

  def get_config(self):
    mi={"gs":self.gs,"k":self.k,"param":self.param,"free":self.free,"self_interaction":self.self_interaction,"self_interaction_const":self.self_interaction_const,"numericalC":self.numericalC,"flag":self.flag,"emptyconst":self.emptyconst}
    th=super(gtopk,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gtopk(**config)
  
  def build(self, input_shape):

    self.metrik=self.add_weight(name="metrik_diag",
                                shape=(self.param,1),
                                initializer=self.metrik_init,#ignores metrik_init completely
                                trainable=self.learnable,
                                regularizer=None)

    self.built=True


  def getmata(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[int(math.floor(i/n)),i]=1
    return ret
  def getmatb(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[i % n,i]=1
    return ret

  def call(self,x):
    x=x[0]

    #print("!x",x.shape)

    gs=self.gs
    k=self.k
    param=self.param
    C=self.numericalC

    #print("gs",gs,"k",k,"param",param,"C",C)


    for i in range(10):t.print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
 
    t.print("calling onetopk",self.metrik, output_stream=sys.stdout)


    #exit()

    mata=K.constant(self.getmata(gs))
    matb=K.constant(self.getmatb(gs))
   
    #print("mata",mata.shape)
    #print("matb",matb.shape) 
    
    xp=K.permute_dimensions(x,(0,2,1))
    #print("xp",xp.shape)

    xa=K.dot(xp,mata)
    xb=K.dot(xp,matb)
    #print("xa",xa.shape,"xb",xb.shape)
    #exit()    

    isval=xa[:,self.flag,:]*xb[:,self.flag,:]

    #return isval,isval
    #print("isval",isval.shape)
    #exit()


    ds=xa-xb
    #print("ds",ds.shape)

    dsp=K.permute_dimensions(ds,(0,2,1))
    #print("dsp",dsp.shape)
    
    dspsq=K.square(dsp)
    #print("dspsq",dspsq.shape)
    #print("self.metrik",self.metrik.shape)
    
    delt=K.dot(dspsq,self.metrik)
    #print("delt",delt.shape)


    delt=K.reshape(delt,(-1,self.gs*self.gs))+(1-isval)*self.emptyconst


    d=K.reshape(delt,(-1,gs,gs))
    #print("d",d.shape)


    #return d,d

    #####no self interactions
    if self.self_interaction==False:
      one=K.eye(gs)
      #print("one",one.shape)
      d+=self.self_interaction_const*one
    #####end no self interactions



    v,_=t.math.top_k(-d,k=k)
    #print("v",v.shape)
    #return v,v
 
    vb=v[:,:,-1]
    #print("vb",vb.shape)

    vbs=K.reshape(vb,(-1,gs,1))
    #print("vbs",vbs.shape)

    
    
    su=d+vbs#plus since top_k(-d)
    #print("su",su.shape)


    #map anything above 0 to 0 and anything below to 1, also map 0 to 1
    #p(-x)=C*d_C(-x)
    #     =d(-C*x)
    #     =1-r(Cx-1)+r(Cx)
    #experimentally:
    #   r(1-Cx)-r(-Cx)

    rel=K.relu(1-C*su)-K.relu(-C*su)
    #print("rel",rel.shape)

    rel=K.relu(rel)-K.relu(rel-1)


    #return rel,rel

    
    dez1=K.reshape(rel,(-1,self.gs*self.gs))
    #print("dez1",dez1.shape)
    dez2=dez1*isval
    #print("dez2",dez2.shape)
    rel=K.reshape(dez2,(-1,self.gs,self.gs))
    print("rel",rel.shape)

    numnei=K.sum(rel,axis=-1)
    print("numnei",numnei.shape)
    factor=self.k/(numnei+0.00000000001)
 
    #return K.concatenate((numnei,factor),axis=-1),factor#,factor#numnei,numnei

    print("factor",factor.shape)
    refactor=K.repeat(factor,self.gs)
    print("refactor",refactor.shape)


    refactor=K.permute_dimensions(refactor,(0,2,1))

    #return refactor,refactor

    rel=rel*refactor 
    print("rel",rel.shape)

    #exit()

    if self.free==0:return rel,x
    zero1=K.zeros_like(x[:,:,0])
    zero1=K.reshape(zero1,(-1,x.shape[1],1))
    #print("!",zero1.shape)
    zerolis=[]
    for i in range(self.free):
      zerolis.append(zero1)
    zeros=K.concatenate(zerolis,axis=-1)
    #print(zeros.shape)

    return rel,K.concatenate((x,zeros),axis=-1)



    



    
  def compute_output_shape(self,input_shape):
    shape=list(input_shape[0])
    #print("inputting",shape,"gs=",self.gs,"sp=",self.param)
    assert len(shape)==3
    assert shape[-1]==self.param
    assert shape[-2]==self.gs

    a1=tuple([shape[0],self.gs,self.gs])
    a2=tuple([shape[0],self.gs,self.param+self.free])

    return a1,a2
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomextractdiag(Layer):#shall take a 5d layer (?,gs,gs,c,c) and take out the diagonal part in gs,gs, resulting in (?,gs,c,c) 
  def __init__(self,gs=20,c=2,**kwargs):
    self.gs=gs
    self.c=c
    super(gcomextractdiag,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                           shape=(self.param,self.c*self.c),
    #                           initializer=self.initializer,
    #                           trainable=self.trainable)


    self.built=True


  def call(self,x):
    x=x[0]#params

    parts=[]
    for i in range(self.gs):
      ac=x[:,i,i,:self.c,:self.c]
      ac=K.reshape(ac,(-1,1,self.c,self.c))
      parts.append(ac)
  
    ret=K.concatenate(parts,axis=1)
    
    return ret


    
    




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==5
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.gs
    assert g_shape[3]==self.c
    assert g_shape[4]==self.c
    return tuple([g_shape[0],self.gs,self.c,self.c])

  def get_config(self):
    mi={"gs":self.gs,"c":self.c}
    th=super(gcomextractdiag,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomextractdiag(**config)
















#NOTWORKINGYET
import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gaddzeros(Layer):
  def __init__(self,inn=20,out=30,param=40,**kwargs):
    assert out>=inn
    self.inn=inn
    self.out=out
    self.param=param
    super(gaddzeros,self).__init__(**kwargs)

  def get_config(self):
    mi={"inn":self.inn,"out":self.out,"param":self.param}
    th=super(gaddzeros,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gaddzeros(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    x=x[0]
    if self.inn==self.out:
      return x
    #print("x",x.shape)
    zero1=K.zeros_like(x[:,0,:])
    #print("zero1",zero1.shape)
    zero1=K.reshape(zero1,(-1,1,x.shape[2]))
    #print("zero1",zero1.shape)
    zerolis=[]
    for i in range(self.out-self.inn):
      zerolis.append(zero1)
    zeros=K.concatenate(zerolis,axis=-2)
    #print("zeros",zeros.shape)
    #print(zeros.shape)

    #exit()

    return K.concatenate((x,zeros),axis=-2)
   

 
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.param
    assert pshape[-2]==self.inn
    pshape[-2]=self.out
    return tuple(pshape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class glom(Layer):#not anymore using keepconst, also not setting the diagonal to zero
  def advrelu(self,x,q):
    i1=(type(q[0])==type(""))
    i2=(type(q[1])==type(""))
    if i1:
      if i2:
        return x
      else:
        return q[1]-K.relu(q[1]-x)
    else:
      if i2:
        return K.relu(x-q[0])+q[0]
      else:
        return K.relu(x-q[0])-K.relu(x-q[1])+q[0]
      
  #high number of iterations fail..why?
  def __init__(self,gs=20,param=40,iterations=1,alinearity=[-1.0,1.0],kernel_initializer='glorot_uniform',self_initializer=None,neig_initializer=None,learnable=True,**kwargs):
    if self_initializer==None:
      self.self_initializer=kernel_initializer
    else:
      self.self_initializer=self_initializer
    if neig_initializer==None:
      self.neig_initializer=kernel_initializer
    else:
      self.neig_initializer=neig_initializer

    self.gs=gs
    self.param=param
    self.iterations=iterations
    self.activate=False
    if len(alinearity)==2:
      self.activate=True
      self.activation=alinearity
    else:
      self.activation=[]
    self.learnable=learnable

    super(glom,self).__init__(**kwargs)

  
  def build(self, input_shape):
    self.neigintact=self.add_weight(name='neigthbourinteraction',#Matrix N
                                shape=(self.param,self.param),
                                initializer=self.neig_initializer,
                                trainable=self.learnable)
    self.selfintact=self.add_weight(name='selfinteraction',#Matrix S
                                shape=(self.param,self.param),
                                initializer=self.self_initializer,
                                trainable=self.learnable)

    self.built=True



  def tp(A,B):
    #A has shape (param,param), B has shape (?,gs,gs), should return (?,gs,gs,param,param)
    return K.zeros((1,self.gs,self.gs,self.param,self.param))#does not do what is intended
    
    





  def call(self,x):
    mat=x[0]#Matrix A
    val=x[1]
   
    N=self.neigintact
    S=self.selfintact
    
    print("N",N.shape)
    print("S",S.shape)
    
    
    t1=self.tp(S,K.eye(self.gs))

    print("t1",t1.shape)



    exit()

 
    for i in range(self.iterations):
      weignei=K.batch_dot(mat,val)#Neighbours of the current nodes
      
      parta=K.dot(weignei,self.neigintact)#Neighbour part
      partb=K.dot(val,self.selfintact)#Self Interaction Part
      val=parta+partb

      if self.activate:
        val=self.advrelu(val,self.activation)

    return val


    
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs

    shape=list(input_shape[1])
    assert len(shape)==3
    assert shape[-1]==self.param
    assert shape[-2]==self.gs

    return tuple(shape)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"iterations":self.iterations,"alinearity":self.activation,"self_initializer":self.self_initializer,"neig_initializer":self.neig_initializer,"learnable":self.learnable}
    th=super(glom,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return glom(**config)













import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomfullyconnected(Layer):#shall take a 3d parameter layer (?,gs,param) and output a fully connected graph of type (?,gs,gs) 
  def __init__(self,gs=20,param=30,**kwargs):
    #c=2
    #mode="min"
    assert gs>0 and param>0
    self.gs=gs
    self.param=param
    super(gcomfullyconnected,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                           shape=(self.param,self.c*self.c),
    #                           initializer=self.initializer,
    #                           trainable=self.trainable)


    self.built=True


  def call(self,x):
    x=x[0]#params  (?,gs,param)
     
    s=x[:,0,0]
    s=K.reshape(s,(-1,1,1))
    s*=0.0
    s+=1.0

    ret=[]
    for i in range(self.gs):
      toa=[]
      for j in range(self.gs):
         toa.append(s)
      ret.append(K.concatenate(toa,axis=1))
    ret=K.concatenate(ret,axis=2)

    return ret 




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==3
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.param
    return tuple([g_shape[0],self.gs,self.gs])

  def get_config(self):
    mi={"gs":self.gs,"param":self.param}
    th=super(gcomfullyconnected,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomfullyconnected(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class glcreate(Layer):
  def __init__(self,gs=20,param=40,**kwargs):
    self.gs=gs
    self.param=param

    super(glcreate,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param}
    th=super(glcreate,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return glcreate(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    #print("!",x[0].shape,x[1].shape)
    mat=x[0]
    val=x[1]
    wei=K.batch_dot(mat,val)
      
    ret=K.concatenate((val,wei),axis=-1)

    
    return ret


    
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs

    shape=list(input_shape[1])
    assert len(shape)==3
    assert shape[-1]==self.param
    assert shape[-2]==self.gs

    return tuple([shape[0],self.gs,self.param*2])
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gecutter(Layer):#like gcutter, but leaves only the last elements
  def __init__(self,inn=30,param=40,out=20,**kwargs):
    assert inn>=out
    self.inn=inn
    self.param=param
    self.out=out
    super(gecutter,self).__init__(**kwargs)


  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]
 
    return x[:,-self.out:,:self.param]

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[-1]==self.param
    assert input_shape[-2]==self.inn
    return tuple([input_shape[0],self.out,self.param])    

  def get_config(self):
    mi={"inn":self.inn,"param":self.param,"out":self.out}
    th=super(gecutter,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gecutter(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gpool(Layer):
  def __init__(self,gs=20,param=40,mode="max",**kwargs):
    self.gs=gs
    self.param=param
    self.mode=mode
    super(gpool,self).__init__(input_shape=(gs,param))

  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]
    if self.mode=="max":
      return K.max(x,axis=-1)
    if self.mode=="mean":
      return K.mean(x,axis=-1)
    if self.mode=="sum":
      return K.sum(x,axis=-1)

    

    
  def compute_output_shape(self,input_shape):
    print("inputting",input_shape,"param",self.param)
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"mode":self.mode}
    th=super(gpool,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpool(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gpoolgrowth(Layer):
  def __init__(self,inn=20,out=30,param=40,kernel_initializer="glorot_uniform",**kwargs):
    self.inn=inn
    self.out=out
    self.param=param
    self.kernel_initializer=kernel_initializer
    super(gpoolgrowth,self).__init__(**kwargs)

  def get_config(self):
    mi={"inn":self.inn,"out":self.out,"param":self.param}
    th=super(gpoolgrowth,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpoolgrowth(**config)
  
  def build(self, input_shape):

    self.inc=self.add_weight(name='inc',
                                shape=(self.param,self.param*(self.out-self.inn)),
                                initializer=self.kernel_initializer,
                                trainable=True)
    self.built=True

  def call(self,x):
    s=x[0]
    add=x[1]

    
    fa=K.dot(s,self.inc)
    fr=K.reshape(fa,(-1,self.out-self.inn,self.param))
    r=K.conatenate((add,fr),axis=-2)
    
    return r
    
     

 
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    xshape=list(input_shape[1])
    assert len(pshape)==2
    assert len(xshape)==3
    assert pshape[-1]==self.param
    assert xshape[-1]==self.param
    assert xshape[-2]==self.inn
    ret=[pshape[0],self.out,self.out]
    return tuple(ret)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class ghealparam(Layer):#takes in a splittet param list (created by gcutparam), and merges them again
  def __init__(self,gs=20,param1=20,param2=20,**kwargs):
    self.gs=gs
    self.param1=param1
    self.param2=param2
    super(ghealparam,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs,"param1":self.param1,"param2":self.param2}
    th=super(ghealparam,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return ghealparam(**config)

  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x1=x[0]
    x2=x[1]
   
    r=K.concatenate((x1,x2),axis=-1)

 
    
    return r

    
  def compute_output_shape(self,input_shape):
    assert len(input_shape)==2
    i1=input_shape[0]
    i2=input_shape[1]
    assert len(i1)==3
    assert len(i2)==3
    assert i1[1]==self.gs
    assert i2[1]==self.gs
    assert i1[2]==self.param1
    assert i2[2]==self.param2
    return tuple([input_shape[0],self.gs,self.param1+self.param2])    


















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomgraphcutter(Layer):#takes a reordered gs*gs graph, and cuts it in pieces of c, which get pooled by a given rule
  def __init__(self,gs=20,c=2,mode="mean",cut=0.5,c_const=1000,**kwargs):
    #c=2
    #mode="min"
    assert gs % c==0#assume there is an equal splitting
    self.gs=gs
    self.c=c
    self.ogs=int(self.gs/self.c)
    self.mode=mode
    self.cut=cut
    self.c_const=c_const
    super(gcomgraphcutter,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                            shape=(self.param*self.c,self.paramo),
    #                            initializer=self.metrik_init,#ignores metrik_init completely
    #                            trainable=self.learnable,
    #                            regularizer=None)



    self.built=True


  def call(self,x):
    A=x[0]#adjacency matrix
 


    Ar=K.reshape(A,(-1,self.ogs,self.c,self.ogs,self.c))
    #print("Ar",Ar.shape)

    if self.mode=="mean":
      Am=K.mean(Ar,axis=(2,4))
    if self.mode=="min":
      Am=K.min(Ar,axis=(2,4))
    if self.mode=="max":
      Am=K.max(Ar,axis=(2,4))

    #print("Am",Am.shape)
    
    C=self.c_const
    cut=self.cut
    
    Ar=K.relu(1+C*(Am-cut))-K.relu(C*(Am-cut))
    

    #print("Ar",Ar.shape)


    #exit()


    return Ar



    exit()

    

    
  def compute_output_shape(self,input_shape):
    grap_shape=input_shape[0]
    assert len(grap_shape)==3
    assert grap_shape[1]==self.gs
    assert grap_shape[2]==self.gs
    return tuple([grap_shape[0],self.ogs,self.ogs])

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"c":self.c,"metrik_init":self.metrik_init,"learnable":self.learnable,"paramo":self.paramo,"mode":self.mode,"cut":self.cut,"c_const":self.c_const}
    th=super(gcomgraphcutter,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomgraphcutter(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gchooseparam(Layer):
  def __init__(self,gs=50,param=30,q=[0,3],**kwargs):
    self.gs=gs
    self.param=param
    self.q=q
    super(gchooseparam,self).__init__(**kwargs)


  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]

    parts=[]
    for q in self.q:
      ac=K.reshape(x[:,:self.gs,q],(-1,self.gs,1))
      parts.append(ac)

    ret=K.concatenate(parts,axis=-1)   

 

    return ret

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[-1]==self.param
    assert input_shape[-2]==self.gs
    return tuple([input_shape[0],self.gs,len(self.q)])    

  def get_config(self):
    mi={"param":self.param,"gs":self.gs,"q":self.q}
    th=super(gchooseparam,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gchooseparam(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace




def metrik_init(shape,dtype=None):
  rel=np.zeros(shape)
  if len(rel)>=4:
    rel[4,4]=1.0
  if len(rel)>=5:
    rel[5,5]=1.0
  return K.constant(rel,dtype=dtype)


class glbuilder(Layer):
  def __init__(self,gs=30,param=40,free=0,metrik_initializer=metrik_init,**kwargs):#param:old parameternumber, free: new parameters initialised to 0
    self.gs=gs
    self.param=param
    self.free=free
    self.metrik_initializer=metrik_initializer
    super(glbuilder,self).__init__(input_shape=(gs,param))
  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"free":self.free}
    th=super(glbuilder,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return glbuilder(**config)
  def getmetrik(self):
    ret=np.zeros((self.param*2,self.param*2))
    ret[4,4]=ret[5,5]=1.0
    return ret  
  def fromdistsq(self,dsq):
    return K.exp(-dsq)

  def getmata(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[int(math.floor(i/n)),i]=1
    return ret
  def getmatb(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[i % n,i]=1
    return ret





  def build(self, input_shape):

    self.metrik=self.add_weight(name='metrik',
                                shape=(self.param,self.param),
                                initializer=self.metrik_initializer,
                                trainable=True)



    self.built=True


  def call(self,x):
    #print(x.shape)

    xt=K.permute_dimensions(x,(0,2,1))


    ca=K.constant(self.getmata(self.gs))
    cb=K.constant(self.getmatb(self.gs))



    #print(ca.shape,cb.shape)

    ma=K.dot(xt,ca)
    mb=K.dot(xt,cb)

    #print(ma.shape,mb.shape)

    ds=ma-mb#?,20,900
    dst=K.permute_dimensions(ds,(0,2,1))#?,900,20
    dsm=K.dot(dst,self.metrik)#?,900,20
    dsa=K.batch_dot(dsm,ds)


    #print(ds.shape,dst.shape,dsm.shape,dsa.shape)

    dsl=t.linalg.diag_part(dsa)

    dsq=K.reshape(dsl,(-1,self.gs,self.gs))


    #print(dsl.shape,dsq.shape)

    #exit()



    #dd4=t.linalg.diag_part(d4)

    #dd4=d4[:,:,1]


    #dsq=K.reshape(dd4,(-1,d,d))

    #print(delta.shape,deltat.shape,metrik.shape,deltatm.shape,d4.shape,dd4.shape,dsq.shape)


    #exit()



    basegraph=self.fromdistsq(dsq)
    
    #print(metrik.shape,xm.shape,xt.shape,dsq.shape)
    #print(basegraph.shape)
    
    parax=x[:,:,:self.param]
    #print(parax.shape)
    
    if self.free==0:return K.concatenate((basegraph,parax),axis=-1) 
    zero1=K.zeros_like(x[:,:,0])
    zero1=K.reshape(zero1,(-1,x.shape[1],1))
    #print("!",zero1.shape)
    zerolis=[]
    for i in range(self.free):
      zerolis.append(zero1)
    zeros=K.concatenate(zerolis,axis=-1)
    #print(zeros.shape)
    
    return K.concatenate((basegraph,parax,zeros),axis=-1)

    
  def compute_output_shape(self,input_shape):
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    return tuple([input_shape[0],self.gs,self.gs+self.param+self.free])    


















import numpy as np
from tensorflow.keras import backend as K
import tensorflow.keras as keras
import tensorflow as t
import json




###definitions of classes that will be used to define a current Network State

class state:
  def __init__(self,gs=10,param=10):
    self.gs=gs
    self.param=param
  def __str__(self):
    return str(self.gs)+"*"+str(self.param)
  def __repr__(self):
    return str(self)
class grap:
  A=None
  X=None
  s=None
  def __init__(self,s):
    self.A=None
    self.X=None
    self.s=s
  def __str__(self):
    return str(self.s)+"("+str(self.A.shape)+"*"+str(self.X.shape)+")"
  def __repr__(self):
    return str(self)
class setting:

  def __init__(self,**kwargs):
    for key in kwargs:
      setattr(self,key,kwargs[key])
  def __str__(self):
    dic={}
    for key in dir(self):
      dic[key]=str(getattr(self,key))
    return json.dumps(dic,indent=2,sort_keys=True)#yes i litterally import json just for this

  def __repr__(self):
    return str(self)





###global attributes that probably will never be chanced

flag=0
self_interaction=True

cut=0.5
c_const=1000.0


###function to create the direct setting object  
def getm():
  m=setting()
  m.usei=False
  m.decompress="classic"

  m.trivial_ladder_n=1
  m.trivial_decompress_activation="linear"
  m.trivial_decompress_init_kernel=t.keras.initializers.TruncatedNormal()
  m.trivial_decompress_init_bias=t.keras.initializers.TruncatedNormal()

  m.sortindex=-1
  m.prenorm=False

  m.graph_init_self=t.keras.initializers.TruncatedNormal()
  m.graph_init_neig=t.keras.initializers.TruncatedNormal()
  m.agraph_init_self=m.graph_init_self
  m.agraph_init_neig=m.graph_init_neig

  m.edges=3#particle net like
  m.edgeactivation="relu"
  m.edgeactivationfinal=m.edgeactivation
  m.edgeusebias=False
  m.edgeconcat=False

  m.gq_activation="relu"
  m.gq_init_kernel=t.keras.initializers.TruncatedNormal()
  m.gq_init_bias=t.keras.initializers.Zeros()
  m.gq_usebias=False
  m.gq_batchnorm=False

  m.shallcomplex=not True
  m.complexsteps=3 

  m.gqa_activation=m.gq_activation
  m.gqa_init_kernel=m.gq_init_kernel
  m.gqa_init_bias=m.gq_init_bias
  m.gqa_usebias=m.gq_usebias
  m.gqa_batchnorm=m.gq_batchnorm

  m.shallacomplex=m.shallcomplex
  m.complexasteps=m.complexsteps

  m.shallredense=False
  m.redenseladder=[8,6]
  m.redenseactivation="relu"
  m.redenseinit=t.keras.initializers.Identity()

  m.compression_init=t.keras.initializers.Identity()

  m.mdense_activation="relu"
  m.mdense_init_kernel=t.keras.initializers.Identity()
  m.mdense_init_bias=t.keras.initializers.Zeros()
  m.mdense_usebias=True
  m.mdense_batchnorm=False



  return m
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gpre5(Layer):#just like gpre2, but with flag at 0
  def __init__(self,gs=20,numericC=10000,**kwargs):
    self.gs=gs
    self.numericC=numericC

    super(gpre5,self).__init__(input_shape=(gs,4))#fixed dimension, since specialised layer

  def get_config(self):
    mi={"gs":self.gs,"numericC":self.numericC}
    th=super(gpre5,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpre5(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    # print("!",x.shape)

    E=K.reshape(x[:,:,0],(-1,self.gs,1))
    p1=K.reshape(x[:,:,1],(-1,self.gs,1))
    p2=K.reshape(x[:,:,2],(-1,self.gs,1))
    p3=K.reshape(x[:,:,3],(-1,self.gs,1))


    
    pt=K.sqrt(p1**2+p2**2)
    p=K.sqrt(pt**2+p3**2)
    iszero=p**2+E**2
    iszero=1-K.relu(1-self.numericC*iszero)+K.relu(-self.numericC*iszero)



    #return iszero

    eta=iszero*0.5*K.log(0.0000000001+(p+p3)/(p-p3+0.0000000001))
    #phi=iszero*t.math.acos(p3/(p+0.0000001))
    phi=iszero*t.math.atan2(p2,p1)

    #print("eta",eta.shape,"phi",phi.shape)


    eta=K.reshape(eta,(-1,self.gs))
    phi=K.reshape(phi,(-1,self.gs))


    meta=K.mean(eta,axis=-1)

    mp1=K.mean(p1,axis=-2)
    mp2=K.mean(p2,axis=-2)

    #print(p2.shape,p1.shape)
    #print(mp1.shape,mp2.shape)
    #exit()

    mphi=t.math.atan2(mp2,mp1)
    #print("meta",meta.shape,"mphi",mphi.shape)

    #mphi=K.mean(phi,axis=-1)

    #exit()


    meta=K.reshape(K.repeat_elements(meta,self.gs,0),(-1,self.gs))
    mphi=K.repeat_elements(mphi,self.gs,1)

    #print("meta",meta.shape,"mphi",mphi.shape)

    #deta=eta#-meta##not sure if abs here
    #dphi=phi#-mphi##not sure here either

    siszero=K.reshape(iszero,(-1,self.gs))

    deta=K.reshape(siszero*(eta-meta),(-1,self.gs,1))
    dphi=K.reshape(siszero*(phi-mphi),(-1,self.gs,1))

    pi=t.constant(math.pi)

    #dphi=K.min([t.math.floormod(dphi,2*pi),t.math.floormod(-dphi,2*pi)],axis=0)

    opta=t.math.floormod(dphi,2*pi)
    optb=t.math.floormod(-dphi,2*pi)

    dphi=t.where(t.greater(opta,optb),optb,-opta)



    #dphi=K.reshape(dphi,(-1,self.gs,1))#should actually be useless?
    
    #dphi=K.min(K.concatenate((t.math.floormod(dphi,2*pi),t.math.floormod(-dphi,2*pi)),axis=-1),axis=-1)

    #dphi=K.reshape(dphi,(-1,self.gs,1))


    #deta=iszero*K.permute_dimensions(K.permute_dimensions(eta,(1,0,2))-meta,(1,0,2))#not sure if abs here required
    #dphi=iszero*K.permute_dimensions(K.permute_dimensions(phi,(1,0,2))-mphi,(1,0,2))#also not sure here either


    spt=K.sum(pt,axis=-2)
    ppt=-iszero*K.permute_dimensions(K.log(0.000000001+K.permute_dimensions(pt,(1,0,2))/(spt+0.0000001)),(1,0,2))#please note the added sign in comparison to the original paper

    #print(iszero.shape,deta.shape,dphi.shape,ppt.shape)
    #print(eta.shape,meta.shape)
    #print(phi.shape,mphi.shape)
    #exit()

    #print(iszero.shape,deta.shape,dphi.shape,pt.shape)
    #exit()


    #meta=K.reshape(meta,(-1,self.gs,1))
    #mphi=K.reshape(mphi,(-1,self.gs,1))
    #phi=K.reshape(phi,(-1,self.gs,1))

    ret=K.concatenate((iszero,deta,dphi,ppt),axis=-1)#adding iszero for numerical reasons
   
    #print(ret.shape,x.shape)
    #exit() 

    return ret

    
  def compute_output_shape(self,input_shape):
    shape=list(input_shape)
    assert len(shape)==3
    assert shape[-1]==4
    assert shape[-2]==self.gs
    shape[-1]=4#?
    return tuple(shape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gkeepbuilder(Layer):
  def __init__(self,gs=30,param=10,free=30,learnable=True,dimension=0,use0=False,**kwargs):
    self.gs=gs
    self.param=param
    self.free=free
    self.learnable=True
    self.dimension=dimension
    self.use0=use0
    super(gkeepbuilder,self).__init__(input_shape=(gs,(gs*(dimension+1))+param))

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"free":self.free,"learnable":self.learnable,"dimension":self.dimension,"use0":self.use0}
    th=super(gkeepbuilder,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gkeepbuilder(**config)


  def fromdistsq(self,dsq):
    return K.exp(-dsq)

  def build(self, input_shape):
    if self.use0:
      self.metrik=self.add_weight(name="weight",
                                shape=(self.dimension+1,1),
                                initializer=keras.initializers.Ones(),
                                trainable=self.learnable)
    else:
      self.metrik=self.add_weight(name="weight",
                                shape=(self.dimension,1),
                                initializer=keras.initializers.Ones(),
                                trainable=self.learnable)
    self.built=True
  def lisshape(self,x):
    ret=[]
    for e in x:
      ret.append(e.shape)
    return ret
  def call(self,x):
    #print(x.shape)
   
    gs=self.gs 
    mats=[]
    for i in range(self.dimension+1):
      amat=(x[:,:gs,gs*i:gs*(i+1)])
      amat=K.reshape(amat,(-1,gs,gs,1))
      mats.append(amat)
    data=x[:,:gs,gs*(self.dimension+1):]

    mat0=mats[0]
    if self.use0:
      mat=K.concatenate(mats,axis=-1)
    else:
      mat=K.concatenate(mats[1:],axis=-1)
    metrik=self.metrik


    
    dsq=K.dot(mat,metrik)
    
    basegraph=self.fromdistsq(dsq)
    #mats.insert(0,basegraph)
    mats[0]=basegraph
    pregraph=K.concatenate(mats,axis=2)
    basegraph=K.reshape(pregraph,(-1,gs,gs*(self.dimension+1))) 
     
    #print(metrik.shape,xm.shape,xt.shape,dsq.shape)
    #print(basegraph.shape)
    
    parax=data
    #print(parax.shape)
    
    if self.free==0:return K.concatenate((basegraph,parax),axis=-1) 
    zero1=K.zeros_like(x[:,:,0])
    zero1=K.reshape(zero1,(-1,x.shape[1],1))
    #print("!",zero1.shape)
    zerolis=[]
    for i in range(self.free):
      zerolis.append(zero1)
    zeros=K.concatenate(zerolis,axis=-1)
    #print(zeros.shape)
    #print("!!!", K.concatenate((basegraph,parax,zeros)).shape,basegraph.shape,pregraph.shape,parax.shape,zeros.shape,self.lisshape(mats)) 
    return K.concatenate((basegraph,parax,zeros))

    
  def compute_output_shape(self,input_shape):
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.gs*(1+self.dimension)+self.param
    return tuple([input_shape[0],input_shape[1],input_shape[2]+self.free])    


















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gvaluation(Layer):
  def __init__(self,gs=20,param=40,metrik_init="glorot_uniform",learnable=True,**kwargs):
    self.gs=gs
    self.param=param
    self.metrik_init=metrik_init
    self.learnable=learnable
    super(gvaluation,self).__init__(**kwargs)

  def build(self, input_shape):


    self.metrik=self.add_weight(name="metrik",
                                shape=(self.param,1),
                                initializer=self.metrik_init,#ignores metrik_init completely
                                trainable=self.learnable,
                                regularizer=None)


    self.built=True


  def call(self,x):
    x=x[0]
 
    #print("x",x.shape)
    
    xp=K.permute_dimensions(x,(0,2,1))
    #print("xp",xp.shape)

    #print("metrik",self.metrik.shape)


    #currently just uses the last value as sorting param
    values=K.reshape(K.dot(x,self.metrik),(-1,self.gs))
    #print("values",values.shape)

    vr=K.reshape(values,(-1,self.gs,1))
    #print("vr",vr.shape)

    ret=K.concatenate((x,vr))
    #print("ret",ret.shape)

    return ret

    exit()

    _,valueorder=t.math.top_k(values,k=self.gs)
    print("valueorder",valueorder.shape)

    xg=t.gather(params=x,indices=valueorder,axis=1,batch_dims=1)
    print("xg",xg.shape)

    xs=K.reshape(xg,(-1,self.ogs,self.param*self.c))
    print("xs",xs.shape)


    traf=K.dot(xs,self.trafo)
    print("traf",traf.shape)

    return traf



    exit()

    

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs,self.param+1])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"metrik_init":self.metrik_init,"learnable":self.learnable}
    th=super(gvaluation,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gvaluation(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gpartinorm(Layer):#takes in -1*gs*param, and returns a array in which for all i,j x[i,:,j] has std 1 (ignores atm that this is kinda stupid for flag)
  def __init__(self,gs=20,param=40,alpha=0.01,**kwargs):
    self.gs=gs
    self.param=param
    self.alpha=alpha#constant to remove numerical problems from std=0
    super(gpartinorm,self).__init__(input_shape=(gs,param))

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"alpha":self.alpha}
    th=super(gpartinorm,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpartinorm(**config)

  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]
    
    #-1,gs,param
    #print("x",x.shape)

    xp=K.permute_dimensions(x,(1,0,2))
    #gs,-1,param
    #print("xp",xp.shape)

    xpr=K.reshape(xp,(self.gs,-1))
    #gs,-1
    #print("xpr",xpr.shape)

    xm0=K.mean(xpr,axis=0)
    xpr-=xm0

    xm=K.mean(K.abs(xpr),axis=0)
    xpr-=xm
        
    xs=K.max(K.abs(xpr),axis=0)+self.alpha
    #-1
    #print("xs",xs.shape)

    xd=xpr/xs
    #gs,-1
    #print("xd",xd.shape)

    xdr=K.reshape(xd,(self.gs,-1,self.param))
    #gs,-1,param
    #print("xdr",xdr.shape)

    xf=K.permute_dimensions(xdr,(1,0,2))
    #-1,gs,param
    #print("xf",xf.shape)

    return xf

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs,self.param])    


















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gfeatkeep(Layer):
  def __init__(self,gs=20,param=40,dimension=0,**kwargs):
    self.gs=gs
    self.param=param
    self.dimension=dimension
    super(gfeatkeep,self).__init__(input_shape=(gs,gs*(dimension+1)+param))

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"dimension":self.dimension}
    th=super(gfeatkeep,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gfeatkeep(**config)

  def build(self, input_shape):

    self.built=True


  def call(self,x):
    return x[:,:,-self.param:]
    

    
  def compute_output_shape(self,input_shape):

    #print("called output of gfeatkeep for is=",input_shape)
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.gs*(self.dimension+1)+self.param
    return tuple([input_shape[0],self.gs,self.param])    


















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gpre1(Layer):
  def __init__(self,gs=20,numericC=10000,**kwargs):
    self.gs=gs
    self.numericC=numericC

    super(gpre1,self).__init__(input_shape=(gs,4))#fixed dimension, since specialised layer

  def get_config(self):
    mi={"gs":self.gs,"numericC":self.numericC}
    th=super(gpre1,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpre1(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    # print("!",x.shape)

    E=K.reshape(x[:,:,0],(-1,self.gs,1))
    p1=K.reshape(x[:,:,1],(-1,self.gs,1))
    p2=K.reshape(x[:,:,2],(-1,self.gs,1))
    p3=K.reshape(x[:,:,3],(-1,self.gs,1))


    
    pt=K.sqrt(p1**2+p2**2)
    p=K.sqrt(pt**2+p3**2)
    iszero=p**2+E**2
    iszero=1-K.relu(1-self.numericC*iszero)+K.relu(-self.numericC*iszero)
    eta=iszero*0.5*K.log(0.0000000001+(p+p3)/(p-p3+0.0000000001))
    #phi=iszero*t.math.acos(p3/(p+0.0000001))
    phi=iszero*t.math.atan2(p2,p1)
    m=K.sqrt(K.relu(E**2-p**2))

    ret= K.concatenate((E,p1,p2,p3,eta,phi,m,pt,p,iszero),axis=-1)
    

    return ret

    
  def compute_output_shape(self,input_shape):
    shape=list(input_shape)
    assert len(shape)==3
    assert shape[-1]==4
    assert shape[-2]==self.gs
    shape[-1]=10#?
    return tuple(shape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomdensediverge(Layer):#shall take a 3d layer (?,gs,param) and output a 4d layer (?,gs,c,paramo)
  def __init__(self,gs=20,param=30,paramo=40,c=2,initializer="glorot_uniform",trainable=True,**kwargs):
    #c=2
    #mode="min"
    self.gs=gs
    self.param=param
    self.paramo=paramo
    self.c=c
    self.initializer=initializer
    self.trainable=trainable
    super(gcomdensediverge,self).__init__(**kwargs)

  def build(self, input_shape):

    self.trafo=self.add_weight(name="trafo",
                               shape=(self.param,self.paramo*self.c),
                               initializer=self.initializer,
                               trainable=self.trainable)


    self.built=True


  def call(self,x):
    x=x[0]#params

    #shall take a 3d feature vector of shape (-1,gs,param), and transform it using trafo into (-1,gs,c,paramo)

    #print("x",x.shape)

    #xi=K.reshape(x,(-1,self.gs,self.ags*self.param))
    #print("xi",xi.shape)

    xt=K.dot(x,self.trafo)
    #print("xt",xt.shape)

    xp=K.reshape(xt,(-1,self.gs,self.c,self.paramo))
    
    #print("xp",xp.shape)

    #exit()

    return xp

    
    




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==3
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.param
    return tuple([g_shape[0],self.gs,self.c,self.paramo])

  def get_config(self):
    mi={"gs":self.gs,"c":self.c,"param":self.param,"paramo":self.paramo,"initializer":self.initializer,"trainable":self.trainable}
    th=super(gcomdensediverge,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomdensediverge(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace

from grapa.layers.kron import kronecker_product_b1 as kron_b1



class gcomgraphlevel(Layer):#takes a (?,gs,gs,c,c) and converts it into (?,c*gs,c*gs)
  def __init__(self,gs=20,c=2,**kwargs):
    self.gs=gs
    self.c=c
    super(gcomgraphlevel,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                           shape=(self.param,self.c*self.c),
    #                           initializer=self.initializer,
    #                           trainable=self.trainable)


    self.built=True

  #def R(self,x):
  #  return K.relu(self.c_const*(x-self.cut)+1)-K.relu(self.c_const*(x-self.c_const))
  def call(self,x):
    x=x[0]
    #print("x",x.shape)
    #exit()
    xp=K.permute_dimensions(x,(0,1,3,2,4))

    ret=K.reshape(xp,(-1,self.gs*self.c,self.gs*self.c))
    

    return ret

    
    




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==5
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.gs
    assert g_shape[3]==self.c
    assert g_shape[4]==self.c
    return tuple([g_shape[0],self.gs*self.c,self.gs*self.c])

  def get_config(self):
    mi={"gs":self.gs,"c":self.c}
    th=super(gcomgraphlevel,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomgraphlevel(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace

from tensorflow.python.ops import array_ops
#from
#https://github.com/tensorflow/tensorflow/blob/r1.8/tensorflow/contrib/kfac/python/ops/utils.py

def kron(mat1, mat2):
  """Computes the Kronecker product two matrices."""
  m1, n1 = mat1.get_shape().as_list()
  mat1_rsh = array_ops.reshape(mat1, [m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [m1 * m2, n1 * n2])
def kron_b1(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1)=1."""
  m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1,m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kron_b2(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat2)=1."""
  m1, n1 = mat1.get_shape().as_list()
  mat1_rsh = array_ops.reshape(mat1, [m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()[1:]
  mat2_rsh = array_ops.reshape(mat2, [-1,1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kron_bb(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1/2)=1."""
  m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1, m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()[1:]
  mat2_rsh = array_ops.reshape(mat2, [-1, 1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,m1 * m2, n1 * n2])
def kron_b1fx1(mat1, mat2):
  """Computes the Kronecker product two matrices, assuming batchdim(mat1)=1."""
  f1, m1, n1 = mat1.get_shape().as_list()[1:]
  mat1_rsh = array_ops.reshape(mat1, [-1,f1,m1, 1, n1, 1])
  m2, n2 = mat2.get_shape().as_list()
  mat2_rsh = array_ops.reshape(mat2, [1, m2, 1, n2])
  return array_ops.reshape(mat1_rsh * mat2_rsh, [-1,f1,m1 * m2, n1 * n2])



#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class ggraphstract(Layer):
  def __init__(self,in1=20,in2=30,**kwargs):
    self.in1=in1
    self.in2=in2
    super(ggraphstract,self).__init__(**kwargs)

  def get_config(self):
    mi={"in1":self.in1,"in2":self.in2}
    th=super(ggraphstract,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return ggraphstract(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    i1=x[0]
    i2=x[1]

    l1=len(i1.shape)
    l2=len(i2.shape)
    


    if l1==3:
      if l2==3:
        return kron_bb(i1,i2)
      else:
        return kron_b1(i1,i2)
    else:
      if l2==3:
        return kron_b2(i1,i2)
      else:
        return kron(i1,i2)


    return kron(i1,i2)

    
    
     

 
  def compute_output_shape(self,input_shape):
    shape1=list(input_shape[0])
    shape2=list(input_shape[1])
    assert len(shape1)==2 or len(shape1)==3
    assert len(shape2)==2 or len(shape2)==3
    assert shape1[-1]==self.in1
    assert shape1[-2]==self.in1
    assert shape2[-1]==self.in2
    assert shape2[-2]==self.in2
    if len(shape1)+len(shape2)==4:
      return tuple([self.in1*self.in2,self.in1*self.in2])
    else:
      return tuple([-1,self.in1*self.in2,self.in1*self.in2])
















import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="grapatf",
    version="0.5",
    author="Simon Klüttermann",
    author_email="Simon.Kluettermann@gmx.de",
    description="A Graph Autoencoder Library for Tensorflow and Keras",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/psorus/grapa/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
	'tensorflow',
	'keras',
	'numpy',
      ],
    download_url='https://github.com/psorus/grapa/archive/0.5.tar.gz',
    
)  



import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomgpool(Layer):
  def __init__(self,gs=20,param=40,paramo=40,c=2,metrik_init="glorot_uniform",learnable=True,mode="mean",cut=0.5,c_const=1000,**kwargs):
    #c=2
    #mode="min"
    assert gs % c==0#assume there is an equal splitting
    self.gs=gs
    self.param=param
    self.metrik_init=metrik_init
    self.c=c
    self.ogs=int(self.gs/self.c)
    self.learnable=learnable
    self.paramo=paramo
    self.mode=mode
    self.cut=cut
    self.c_const=c_const
    super(gcomgpool,self).__init__(**kwargs)

  def build(self, input_shape):

    self.trafo=self.add_weight(name="trafo",
                                shape=(self.param*self.c,self.paramo),
                                initializer=self.metrik_init,#ignores metrik_init completely
                                trainable=self.learnable,
                                regularizer=None)

#    self.metrik=self.add_weight(name="metrik",
#                                shape=(self.param,1),
#                                initializer=keras.initializers.ones(),#ignores metrik_init completely
#                                trainable=not self.learnable,
#                                regularizer=None)


    self.built=True


  def call(self,x):
    A=x[0]#adjacency matrix
    x=x[1]#parameters
 
    #print("x",x.shape)
    #print("A",A.shape) 

    #print("trafo",self.trafo.shape)


    #currently just uses the last value as sorting param
    values=x[:,:,-1]#K.reshape(K.dot(x,self.metrik),(-1,self.gs))
    #print("values",values.shape)

    _,valueorder=t.math.top_k(values,k=self.gs)
    #print("valueorder",valueorder.shape)

    #valueorder=t.argsort(values,axis=-1)
    #print("valueorder",valueorder.shape)

    xg=t.gather(params=x,indices=valueorder,axis=1,batch_dims=1)
    #print("xg",xg.shape)

    #exit()

    xs=K.reshape(xg,(-1,self.ogs,self.param*self.c))
    #print("xs",xs.shape)


    traf=K.dot(xs,self.trafo)
    #print("traf",traf.shape)



    At1=t.gather(params=A,indices=valueorder,axis=1,batch_dims=1)
    At2=t.gather(params=At1,indices=valueorder,axis=2,batch_dims=1)

    #print("At1",At1.shape,"At2",At2.shape)


    Ar=K.reshape(At2,(-1,self.ogs,self.c,self.ogs,self.c))
    #print("Ar",Ar.shape)

    if self.mode=="mean":
      Am=K.mean(Ar,axis=(2,4))
    if self.mode=="min":
      Am=K.min(Ar,axis=(2,4))
    if self.mode=="max":
      Am=K.max(Ar,axis=(2,4))

    #print("Am",Am.shape)
    
    C=self.c_const
    cut=self.cut
    
    Ar=K.relu(1+C*(Am-cut))-K.relu(C*(Am-cut))
    

    #print("Ar",Ar.shape)


    #exit()


    return Ar,traf



    exit()

    

    
  def compute_output_shape(self,input_shape):
    grap_shape=input_shape[0]
    input_shape=input_shape[1]
    assert len(input_shape)==3
    assert len(grap_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    assert grap_shape[1]==self.gs
    assert grap_shape[2]==self.gs
    return tuple([grap_shape[0],self.ogs,self.ogs]),tuple([input_shape[0],self.ogs,self.paramo])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"c":self.c,"metrik_init":self.metrik_init,"learnable":self.learnable,"paramo":self.paramo,"mode":self.mode,"cut":self.cut,"c_const":self.c_const}
    th=super(gcomgpool,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomgpool(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gltknd(Layer):
  def advrelu(self,x,q):
    i1=(type(q[0])==type(""))
    i2=(type(q[1])==type(""))
    if i1:
      if i2:
        return x
      else:
        return q[1]-K.relu(q[1]-x)
    else:
      if i2:
        return K.relu(x-q[0])+q[0]
      else:
        return K.relu(x-q[0])-K.relu(x-q[1])+q[0]
      
  #high number of iterations fail..why?
  def __init__(self,gs=20,param=40,keepconst=10,iterations=1,alinearity=[-1.0,1.0],kernel_initializer='glorot_uniform',self_initializer=None,neig_initializer=None,learnable=True,**kwargs):
    if self_initializer==None:
      self.self_initializer=kernel_initializer
    else:
      self.self_initializer=self_initializer
    if neig_initializer==None:
      self.neig_initializer=kernel_initializer
    else:
      self.neig_initializer=neig_initializer

    self.gs=gs
    self.param=param
    self.keepconst=keepconst
    self.makezerolmat=K.constant(self.genmakezerolmat(gs))
    self.iterations=iterations
    self.activate=False
    if len(alinearity)==2:
      self.activate=True
      self.activation=alinearity#most general form of continous activation: const,x,const
    else:
      self.activation=[]
    self.learnable=learnable

    super(gltknd,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"keepconst":self.keepconst,"iterations":self.iterations,"alinearity":self.activation,"self_initializer":self.self_initializer,"neig_initializer":self.neig_initializer,"learnable":self.learnable}
    th=super(gltknd,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gltknd(**config)
  
  def build(self, input_shape):
    self.neigintact=self.add_weight(name='neigthbourinteraction',
                                shape=(self.param,self.param-self.keepconst,),
                                initializer=self.neig_initializer,
                                trainable=self.learnable)
    self.selfintact=self.add_weight(name='selfinteraction',
                                shape=(self.param,self.param-self.keepconst,),
                                initializer=self.self_initializer,
                                trainable=self.learnable)

    self.built=True

  def genmakezerolmat(self,n):
    ret=np.zeros((n,n))+1.0
    for i in range(n):
      ret[i,i]=0.0
    return ret

  def call(self,x):
    #print("!",x[0].shape,x[1].shape)
    mat=x[0]
    val=x[1]
    con=val[:,:,:self.keepconst]
    var=val[:,:,self.keepconst:]
    mat0=mat*self.makezerolmat
    
    tra=trace(mat)
    sumtra=K.sum(mat,axis=-1)
    msumtra=K.mean(sumtra,axis=-1)

    #print("tra",tra.shape)
    #print("sumtra",sumtra.shape)

    #exit()

    # print(mat0.shape,val.shape)
    # val=K.permute_dimensions(val,(1,0,2))
    # print(mat0.shape,val.shape)
    # mat0=K.permute_dimensions(mat0,(1,0,2))
    # val =K.permute_dimensions(val ,(1,2,0))
    # print(mat0.shape,val.shape)

    # weignei=K.dot(mat0,val)
    # weignei=K.dot(val,mat0)
    for i in range(self.iterations):
      weignei=K.batch_dot(mat0,val)
      
      # print("---",weignei.shape)
      # return K.sum(K.sum(weignei,axis=-1),axis=-1)
      # exit()
      
      # print(weignei.shape)
      
      
      parta=K.dot(weignei,self.neigintact)
      partb=K.dot(val,self.selfintact)
      # print(parta.shape,tra.shape)

      # print(parta.shape)
      
      # exit()
      var=parta+partb

      #var=K.permute_dimensions(var,(1,2,0))
      
      # return tra
      
      #var/=msumtra
      #var=K.permute_dimensions(var,(2,0,1))

      if self.activate:
        var=self.advrelu(var,self.activation)


      val=K.concatenate((con,var),axis=-1)

      # return K.sum(K.sum(val,axis=-1),axis=-1)
      # print("###",K.eval(val))
      
      
      # print(parta.shape,partb.shape,var.shape,val.shape)
    
    
    # exit()
    return val
    #return K.concatenate((mat,val),axis=-1)
    
    # print("call")
    # # print(x.shape)
    # m=((K.dot(K.pow(x,2),self.kernel)))
    # # print("!",m.shape)
    # E=((K.dot(x,K.constant(np.array([[1.0],[0.0],[0.0],[0.0]])))))
    # px=((K.dot(x,K.constant(np.array([[0.0],[1.0],[0.0],[0.0]])))))
    # py=((K.dot(x,K.constant(np.array([[0.0],[0.0],[1.0],[0.0]])))))
    # pt=K.sqrt(K.square(px)+K.square(py))


    
  def compute_output_shape(self,input_shape):
    # return tuple([input_shape[0],40,60])
    # return tuple([input_shape[0]])
    # return tuple(input_shape)
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs

    shape=list(input_shape[1])
    assert len(shape)==3
    assert shape[-1]==self.param
    assert shape[-2]==self.gs
    #shape[-1]=self.graphmax+self.graphvar#this layer should not chance the size of the network, so this line becomes kinda useless
    # shape[-2]=self.graphmax
    return tuple(shape)
    # return tuple([15,2])
    # return tuple(K.constant([1,1]).shape)
    
#    assert len(shape)==2
    # assert shape[-1]==4
    # shape[-1]=2
    # return tuple(shape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomreopool(Layer):
  def __init__(self,gs=20,param=40,**kwargs):
    #c=2
    #mode="min"
    self.gs=gs
    self.param=param
    super(gcomreopool,self).__init__(**kwargs)

  def build(self, input_shape):



    self.built=True


  def call(self,x):
    A=x[0]#adjacency matrix
    x=x[1]#parameters
 
    #print("x",x.shape)
    #print("A",A.shape) 

    #print("trafo",self.trafo.shape)


    #currently just uses the last value as sorting param
    values=x[:,:,-1]#K.reshape(K.dot(x,self.metrik),(-1,self.gs))
    #print("values",values.shape)

    _,valueorder=t.math.top_k(values,k=self.gs)
    #print("valueorder",valueorder.shape)

    #valueorder=t.argsort(values,axis=-1)
    #print("valueorder",valueorder.shape)

    xg=t.gather(params=x,indices=valueorder,axis=1,batch_dims=1)
    #print("xg",xg.shape)

    #exit()

    #xs=K.reshape(xg,(-1,self.ogs,self.param*self.c))
    #print("xs",xs.shape)


    #traf=K.dot(xs,self.trafo)
    #print("traf",traf.shape)



    At1=t.gather(params=A,indices=valueorder,axis=1,batch_dims=1)
    At2=t.gather(params=At1,indices=valueorder,axis=2,batch_dims=1)

    return At2,xg



    exit()

    

    
  def compute_output_shape(self,input_shape):
    grap_shape=input_shape[0]
    input_shape=input_shape[1]
    assert len(input_shape)==3
    assert len(grap_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    assert grap_shape[1]==self.gs
    assert grap_shape[2]==self.gs
    return tuple([grap_shape[0],self.gs,self.gs]),tuple([input_shape[0],self.gs,self.param])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param}
    th=super(gcomreopool,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomreopool(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gmultiply(Layer):
  def __init__(self,gs=30,param=40,c=2,**kwargs):
    self.gs=gs
    self.param=param
    self.c=c
    super(gmultiply,self).__init__(**kwargs)


  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]
   



    l=[]
    for i in range(self.c):
      l.append(x)

    #print("l",l,len(l),self.c)
    #exit()




    q=K.concatenate(l,axis=-1)
    #print("q",q.shape)
    #exit()
    ret=K.reshape(q,(-1,self.gs*self.c,self.param))
    return ret
 
    
    exit() 



    return x

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[-1]==self.param
    assert input_shape[-2]==self.gs
    return tuple([input_shape[0],self.gs*self.c,self.param])    

  def get_config(self):
    mi={"ga":self.gs,"param":self.param,"c":self.c}
    th=super(gmultiply,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gmultiply(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace
from grapa.layers.kron import kronecker_product as kron
from grapa.layers.kron import kronecker_product_b1 as kron_b1
from grapa.layers.kron import kronecker_product_b1fx1 as kron_b1fx1
from grapa.layers.kron import kronecker_product_b2 as kron_b2




#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class glam(Layer):#not anymore using keepconst, also not setting the diagonal to zero
  def advrelu(self,x,q):
    i1=(type(q[0])==type(""))
    i2=(type(q[1])==type(""))
    if i1:
      if i2:
        return x
      else:
        return q[1]-K.relu(q[1]-x)
    else:
      if i2:
        return K.relu(x-q[0])+q[0]
      else:
        return K.relu(x-q[0])-K.relu(x-q[1])+q[0]
      
  #high number of iterations fail..why?
  def __init__(self,gs=20,param=40,a=10,iterations=1,alinearity=[-1.0,1.0],kernel_initializer='glorot_uniform',self_initializer=None,neig_initializer=None,learnable=True,**kwargs):
    if self_initializer==None:
      self.self_initializer=kernel_initializer
    else:
      self.self_initializer=self_initializer
    if neig_initializer==None:
      self.neig_initializer=kernel_initializer
    else:
      self.neig_initializer=neig_initializer

    self.gs=gs
    self.param=param
    self.a=a
    self.iterations=iterations
    self.activate=False
    if len(alinearity)==2:
      self.activate=True
      self.activation=alinearity
    else:
      self.activation=[]
    self.learnable=learnable

    super(glam,self).__init__(**kwargs)

  
  def build(self, input_shape):
    self.neigintact=self.add_weight(name='neigthbourinteraction',#Matrix N
                                shape=(self.param,self.param),
                                initializer=self.neig_initializer,
                                trainable=self.learnable)
    self.selfintact=self.add_weight(name='selfinteraction',#Matrix S
                                shape=(self.param,self.param),
                                initializer=self.self_initializer,
                                trainable=self.learnable)

    self.built=True

  def call(self,x):
    mat=x[0]#Matrix A
    val=x[1]

    eye=K.eye(self.gs)
 
    p1=kron(eye,self.selfintact)
    p2=kron_b1fx1(mat,self.neigintact)

    p=p1+p2


 

    v=K.reshape(val,(-1,self.gs*self.param))
    p=K.reshape(p,(-1,self.gs*self.param,self.gs*self.param))

    #print("p",p.shape,"v",v.shape)
    #exit()
    #does not yet work, how to define products with 2 batch dimensions
    #a bit inelegant if you ask me   

    for i in range(self.iterations):
      v=K.batch_dot(p,v)
      if self.activate:
        v=self.advrelu(v,self.activation)
    #print("v",v.shape,"a",self.a,"gs",self.gs,"param",self.param)
    #exit()
 
    ret=K.reshape(v,(-1,self.a,self.gs,self.param))#keine Ahnung ob richtig rum #doch scheint zu passen

    return ret




    #    v=K.reshape(val,(-1,self.a,self.gs*self.param))
    #    #p=K.reshape(p,(-1,self.gs*self.param,self.gs*self.param))
    #
    #    #print("p",p.shape,"v",v.shape)
    #    #exit()
    #    #does not yet work, how to define products with 2 batch dimensions
    #    #a bit inelegant if you ask me   
    #
    #    vs=[]
    #    for j in range(self.a):
    #      av=v[:,j,:]
    #      ap=p[:,j,:,:]
    #      for i in range(self.iterations):
    #        av=K.batch_dot(ap,av)
    #        if self.activate:
    #          av=self.advrelu(av,self.activation)
    #      av=K.reshape(av,(-1,1,self.gs*self.param))
    #      vs.append(av)
    #    v=K.concatenate(vs,axis=1)
    #    #print("v",v.shape,"a",self.a,"gs",self.gs,"param",self.param)
    #    #exit()
    # 
    #    ret=K.reshape(v,(-1,self.a,self.gs,self.param))#keine Ahnung ob richtig rum #doch scheint zu passen
    #
    #    return ret






    
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==4
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs
    assert pshape[-3]==self.a

    shape=list(input_shape[1])
    assert len(shape)==4
    assert shape[-1]==self.param
    assert shape[-2]==self.gs
    assert shape[-3]==self.a

    return tuple(shape)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"a":self.a,"iterations":self.iterations,"alinearity":self.activation,"self_initializer":self.self_initializer,"neig_initializer":self.neig_initializer,"learnable":self.learnable}
    th=super(glam,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return glam(**config)













import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace


def R(self,x,c_const=1000.0,cut=0.5):
  return K.relu(c_const*(x-cut)+1)-K.relu(c_const*(x-c_const))
def doand(x,mode="prod",c_const=1000.0,cut=0.5):


  if mode=="and":
    ret=R(K.prod(x,axis=-1),c_const=c_const,cut=cut)
  if mode=="prod":
    ret=K.prod(x,axis=-1)
  if mode=="or":
    ret=R(K.sum(x,axis=-1),c_const=c_const,cut=cut)
  if mode=="sum":
    ret=K.sum(x,axis=-1)


  return ret



class gcomgraphand2(Layer):#is capable of running "and" operations on two (?,c*gs,c*gs) graphs, resulting in (?,c*gs,c*gs) 
  def __init__(self,gs=20,c=2,mode="prod",cut=0.5,c_const=1000.0,**kwargs):
    self.gs=gs
    self.c=c
    self.mode=mode
    self.cut=cut
    self.c_const=c_const
    super(gcomgraphand2,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                           shape=(self.param,self.c*self.c),
    #                           initializer=self.initializer,
    #                           trainable=self.trainable)


    self.built=True

  def call(self,q):
    y=q[1]
    x=q[0]

    x=K.reshape(x,(-1,self.gs*self.c,self.gs*self.c,1))
    y=K.reshape(y,(-1,self.gs*self.c,self.gs*self.c,1))

    q=K.concatenate((x,y),axis=-1)

    ret=doand(q,mode=self.mode,c_const=self.c_const,cut=self.cut)
    
    return ret

    
    




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    g_shape2=input_shape[1]
    assert len(g_shape)==3
    assert g_shape[1]==self.gs*self.c
    assert g_shape[2]==self.gs*self.c
    assert g_shape2[1]==self.gs*self.c
    assert g_shape2[2]==self.gs*self.c
    return tuple([g_shape[0],self.gs*self.c,self.gs*self.c])

  def get_config(self):
    mi={"gs":self.gs,"c":self.c,"mode":self.mode,"cut":self.cut,"c_const":self.c_const}
    th=super(gcomgraphand2,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomgraphand2(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gkeepcutter(Layer):
  def __init__(self,inn=30,param=40,out=20,dimension=0,**kwargs):
    assert inn>=out
    self.inn=inn
    self.param=param
    self.out=out
    self.dimension=dimension
    super(gkeepcutter,self).__init__(input_shape=(inn,inn*(dimension+1)+param))


  def build(self, input_shape):

    self.built=True


  def call(self,x):
    
    x=x[:,:self.out,:]
    mainmat=x[:,:,:self.out]
    otmats=[]
    for i in range(self.dimension):
      otmats.append(x[:,:,self.inn*(i+1):self.inn*(i+1)+self.out])
    params=x[:,:,-self.param:]
    
    otmats.insert(0,mainmat)
    otmats.append(params)

    return K.concatenate(otmats,axis=-1)

    
  def compute_output_shape(self,input_shape):
    assert len(input_shape)==3
    assert input_shape[1]==self.inn
    assert input_shape[2]==self.inn*(self.dimension+1)+self.param
    return tuple([input_shape[0],self.out,self.out*(self.dimension+1)+self.param])    

  def get_config(self):
    mi={"inn":self.inn,"param":self.param,"out":self.out,"dimension":self.dimension}
    th=super(gkeepcutter,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gkeepcutter(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense,Activation
from tensorflow.keras.layers import BatchNormalization as BatchNorm
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gltrivmlp(Layer):
  def advrelu(self,x,q):
    i1=(type(q[0])==type(""))
    i2=(type(q[1])==type(""))
    if i1:
      if i2:
        return x
      else:
        return q[1]-K.relu(q[1]-x)
    else:
      if i2:
        return K.relu(x-q[0])+q[0]
      else:
        return K.relu(x-q[0])-K.relu(x-q[1])+q[0]
      
  #high number of iterations fail..why?
  def __init__(self,gs=20,param=40,keepconst=10,iterations=10,alinearity=[-1.0,1.0],initializer='glorot_uniform',i1=30,i2=20,mlpact=K.relu,momentum=0.99,k=16,**kwargs):
    self.initializer=initializer
    self.gs=gs
    self.param=param
    self.keepconst=keepconst
    self.iterations=iterations
    self.activate=False
    self.i1=i1
    self.i2=i2
    self.mlpact=mlpact
    self.momentum=momentum
    self.k=k



    if len(alinearity)==2:
      self.activate=True
      self.activation=alinearity#most general form of continous activation: const,x,const
    else:
      self.activation=[]

    super(gltrivmlp,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"keepconst":self.keepconst,"iterations":self.iterations,"alinearity":self.activation,"initializer":self.initializer,"i1":self.i1,"i2":self.i2,"mlpact":self.mlpact,"momentum":self.momentum,"k":self.k}
    th=super(gltrivmlp,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gltrivmlp(**config)

  def getmata(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[int(math.floor(i/n)),i]=1
    return ret
  def getmatb(self,n):
    ret=np.zeros((n,n*n))
    for i in range(n*n):
      ret[i % n,i]=1
    return ret



  def mlp(self,q):#does the mlp, should always be build, so that q can be arbitrary dimensional, als long as q.shape[-1]=2*gs, and return.shape=q.shape, with the exception of return.shape[-1]=gs-keepconst,this cannot be (easily) reached thanks to the batch layers

    #print("q",q.shape)
    #exit()

    return q[:,:,:self.param-self.keepconst]
    return K.dot(q,self.tmat)


  def build(self, input_shape):
    self.tmat=self.add_weight(name='tmat',
                                shape=(2*self.param,self.param-self.keepconst,),
                                initializer=self.initializer,
                                trainable=False)
    
    self.built=True


  def call(self,x):
    #print("!",x[0].shape,x[1].shape)
    mat=x[0]
    val=x[1]
    con=val[:,:,:self.keepconst]
    var=val[:,:,self.keepconst:]
    

    mata=K.constant(self.getmata(self.gs))
    matb=K.constant(self.getmatb(self.gs))

    for i in range(self.iterations):

      #print("val",val.shape)

      valp=K.permute_dimensions(val,(0,2,1))
      
      #print("valp",valp.shape)

      vA=K.dot(valp,mata)
      vB=K.dot(valp,matb)

      #print("vA",vA.shape,"vB",vB.shape)

      feat=K.permute_dimensions(vA,(0,2,1))
      diff=K.permute_dimensions(vB-vA,(0,2,1))
      
      #print("feat",feat.shape,"diff",diff.shape)

      premlp=K.concatenate((feat,diff),axis=-1)

      #print("premlp",premlp.shape)

      postmlp=premlp[:,:,:self.param-self.keepconst]
      #postmlp=self.mlp(premlp)
      
      #print("postmlp",postmlp.shape)

      ppmlp=K.permute_dimensions(postmlp,(0,2,1))

      #print("ppmlp",ppmlp.shape)
      
      res=K.reshape(ppmlp,(-1,self.param-self.keepconst,self.gs,self.gs))

      #print("res",res.shape)
      #print("mat",mat.shape)

      resp=K.permute_dimensions(res,(1,0,2,3)) 


      presmat=resp*mat#kinda wondering that this(resp*mat) actually works...tja it actually does not

      #print("presmat",presmat.shape)

      resmat=K.permute_dimensions(presmat,(1,0,2,3))
      #resmat=K.permute_dimensions(presmat,(1,2,0,3))

      print("resmat",resmat.shape)

      #exit()


      summ=K.sum(resmat,axis=-1)#/msumtra

      print("summ",summ.shape)


      #print("summ",summ.shape)

      #print("pre",summ.shape)
      summ/=self.k
      #print("post",summ.shape)

      #exit()
     

      #print("permuted from",summ.shape) 
      
      #var=K.permute_dimensions(summ,(2,0,1))

      var=K.permute_dimensions(summ,(0,2,1))#hopefully implemented this into the resmat permute, nope, not diffbar

      #print("permuted to",var.shape)


      print("var",var.shape)
      #print("con",con.shape)

      #exit()

      if self.activate:
        var=self.advrelu(var,self.activation)


      val=K.concatenate((con,var),axis=-1)

      #print("concatting",con.shape,var.shape,"=>",val.shape)
      #print("val",val.shape)       

      continue

    #print("returning",val.shape)
    #exit()
 
    return val
    #return K.concatenate((mat,val),axis=-1)
    
    # print("call")
    # # print(x.shape)
    # m=((K.dot(K.pow(x,2),self.kernel)))
    # # print("!",m.shape)
    # E=((K.dot(x,K.constant(np.array([[1.0],[0.0],[0.0],[0.0]])))))
    # px=((K.dot(x,K.constant(np.array([[0.0],[1.0],[0.0],[0.0]])))))
    # py=((K.dot(x,K.constant(np.array([[0.0],[0.0],[1.0],[0.0]])))))
    # pt=K.sqrt(K.square(px)+K.square(py))


    
  def compute_output_shape(self,input_shape):
    # return tuple([input_shape[0],40,60])
    # return tuple([input_shape[0]])
    # return tuple(input_shape)
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs

    shape=list(input_shape[1])
    assert len(shape)==3
    assert shape[-1]==self.param
    assert shape[-2]==self.gs
    #shape[-1]=self.graphmax+self.graphvar#this layer should not chance the size of the network, so this line becomes kinda useless
    # shape[-2]=self.graphmax
    return tuple(shape)
    # return tuple([15,2])
    # return tuple(K.constant([1,1]).shape)
    
#    assert len(shape)==2
    # assert shape[-1]==4
    # shape[-1]=2
    # return tuple(shape)
















#does not work

import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomdex(Layer):#creates ordering by last param (returns indices)
  def __init__(self,gs=20,param=40,**kwargs):
    self.gs=gs
    self.param=param
    super(gcomdex,self).__init__(**kwargs)

  def build(self, input_shape):


    self.built=True


  def call(self,x):
    x=x[0]
 
    values=x[:,:,-1]#K.reshape(K.dot(x,self.metrik),(-1,self.gs))

    _,valueorder=t.math.top_k(values,k=self.gs)

    valueorder=K.cast(valueorder,"float32")

    return valueorder

    

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param}
    th=super(gcomdex,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomdex(**config)













